@@ -20,6 +20,7 @@
  ***************************************************************************/
 #include <algorithm>
 #include <util/file.h>
+#include <util/array.h>
 #include <qstringlist.h>
 #include "chunkmanager.h"
 #include "torrent.h"
@@ -29,6 +30,7 @@
 #include "singlefilecache.h"
 #include "multifilecache.h"
 #include <util/log.h>
+#include <util/functions.h>
 #include "globals.h"
 
 #include <klocale.h>
@@ -36,6 +38,7 @@
 namespace bt
 {
 	
+	Uint32 ChunkManager::max_chunk_size_for_data_check = 0;
 	
 
 	ChunkManager::ChunkManager(
@@ -44,8 +47,11 @@ namespace bt
 			const QString & datadir,
 			bool custom_output_name)
 	: tor(tor),chunks(tor.getNumChunks()),
-	bitset(tor.getNumChunks()),excluded_chunks(tor.getNumChunks())
+	bitset(tor.getNumChunks()),excluded_chunks(tor.getNumChunks()),only_seed_chunks(tor.getNumChunks()),todo(tor.getNumChunks())
 	{
+		during_load = false;
+		only_seed_chunks.setAll(false);
+		todo.setAll(true);
 		if (tor.isMultiFile())
 			cache = new MultiFileCache(tor,tmpdir,datadir,custom_output_name);
 		else
@@ -68,21 +74,17 @@ namespace bt
 		chunks.setAutoDelete(true);
 		chunks_left = 0;
 		recalc_chunks_left = true;
+		corrupted_count = recheck_counter = 0;
 
 		for (Uint32 i = 0;i < tor.getNumFiles();i++)
 		{
 			TorrentFile & tf = tor.getFile(i);
-			connect(&tf,SIGNAL(downloadStatusChanged(TorrentFile*, bool )),
-					 this,SLOT(downloadStatusChanged(TorrentFile*, bool )));
-			connect(&tf,SIGNAL(downloadPriorityChanged(TorrentFile*, Priority )),
-					 this,SLOT(downloadPriorityChanged(TorrentFile*, Priority )));
-			if (tf.doNotDownload())
+			connect(&tf,SIGNAL(downloadPriorityChanged(TorrentFile*, Priority, Priority )),
+					 this,SLOT(downloadPriorityChanged(TorrentFile*, Priority, Priority )));
+			
+			if (tf.getPriority() != NORMAL_PRIORITY)
 			{
-				downloadStatusChanged(&tf,false);
-			}
-			if (tf.getPriority() == FIRST_PRIORITY || tf.getPriority() == LAST_PRIORITY)
-			{
-				downloadPriorityChanged(&tf,tf.getPriority());
+				downloadPriorityChanged(&tf,tf.getPriority(),tf.getOldPriority());
 			}
 		}
 	
@@ -91,7 +93,7 @@ namespace bt
 			for(Uint32 i=0; i<tor.getNumFiles(); ++i)
 			{
 				bt::TorrentFile & file = tor.getFile(i);
-				if(file.isMultimedia()) 
+				if (file.isMultimedia() && file.getPriority() != bt::ONLY_SEED_PRIORITY) 
 				{
 					prioritise(file.getFirstChunk(), file.getFirstChunk()+1, PREVIEW_PRIORITY);
 					if (file.getLastChunk() - file.getFirstChunk() > 2)
@@ -101,7 +103,7 @@ namespace bt
 		}
 		else
 		{
-			if(tor.isMultimedia() )
+			if(tor.isMultimedia())
 			{
 				prioritise(0,1,PREVIEW_PRIORITY);
 				if (tor.getNumChunks() > 2)
@@ -124,36 +126,15 @@ namespace bt
 
 	void ChunkManager::changeDataDir(const QString & data_dir)
 	{
-		QValueList<Uint32> mapped;
-		// save all buffered and mapped chunks
-		for (Uint32 i = 0;i < tor.getNumChunks();i++)
-		{
-			Chunk* c = getChunk(i);
-			if (c->getStatus() == Chunk::MMAPPED ||
-				c->getStatus() == Chunk::BUFFERED)
-			{
-				cache->save(c);
-				mapped.append(i);
-			}
-		}
-		cache->close();
 		cache->changeTmpDir(data_dir);
-		cache->open();
-		// reload the previously mapped and buffered chunks
-		for (Uint32 i = 0;i < mapped.count();i++)
-		{
-			Chunk* c = getChunk(mapped[i]);
-			cache->load(c);
-		}
-		
 		index_file = data_dir + "index";
 		file_info_file = data_dir + "file_info";
 		file_priority_file = data_dir + "file_priority";
-		savePriorityInfo();
 	}
 	
 	void ChunkManager::loadIndexFile()
 	{
+		during_load = true;
 		loadPriorityInfo();
 		
 		File fptr;
@@ -162,6 +143,7 @@ namespace bt
 			// no index file, so assume it's empty
 			bt::Touch(index_file,true);
 			Out(SYS_DIO|LOG_IMPORTANT) << "Can't open index file : " << fptr.errorString() << endl;
+			during_load = false;
 			return;
 		}
 
@@ -178,11 +160,13 @@ namespace bt
 				{
 					c->setStatus(Chunk::ON_DISK);
 					bitset.set(hdr.index,true);
+					todo.set(hdr.index,false);
 					recalc_chunks_left = true;
 				}
 			}
 		}
 		tor.updateFilePercentage(bitset);
+		during_load = false;
 	}
 	
 	void ChunkManager::saveIndexFile()
@@ -267,9 +251,37 @@ namespace bt
 		{
 			// load the chunk if it is on disk
 			cache->load(c);
-			loaded.append(i);
+			loaded.insert(i,bt::GetCurrentTime());
+			bool check_allowed = (max_chunk_size_for_data_check == 0 || tor.getChunkSize() <= max_chunk_size_for_data_check);
+			
+			// when no corruptions have been found, only check once every 5 chunks
+			if (check_allowed && recheck_counter < 5 && corrupted_count == 0)
+				check_allowed = false; 
+			 
+			if (c->getData() && check_allowed)
+			{
+				recheck_counter = 0;
+				if (!c->checkHash(tor.getHash(i)))
+				{
+					Out(SYS_DIO|LOG_IMPORTANT) << "Chunk " << i 
+							<< " has been found invalid, redownloading" << endl;
+				
+					resetChunk(i);
+					tor.updateFilePercentage(i,bitset);
+					saveIndexFile();
+					recalc_chunks_left = true;
+					corrupted_count++;
+					corrupted(i);
+					return 0;
+				}
+			}
+			else
+			{
+				recheck_counter++;
+			}
 		}
 		
+		loaded.insert(i,bt::GetCurrentTime());
 		return c;
 	}
 		
@@ -300,6 +312,7 @@ namespace bt
 		c->clear();
 		c->setStatus(Chunk::NOT_DOWNLOADED);
 		bitset.set(i,false);
+		todo.set(i,!excluded_chunks.get(i) && !only_seed_chunks.get(i));
 		loaded.remove(i);
 		tor.updateFilePercentage(i,bitset);
 	}
@@ -307,17 +320,20 @@ namespace bt
 	void ChunkManager::checkMemoryUsage()
 	{
 		Uint32 num_removed = 0;
-		QValueList<Uint32>::iterator i = loaded.begin();
+		QMap<Uint32,TimeStamp>::iterator i = loaded.begin();
 		while (i != loaded.end())
 		{
-			Chunk* c = chunks[*i];
-			if (!c->taken())
+			Chunk* c = chunks[i.key()];
+			// get rid of chunk if nobody asked for it in the last 5 seconds
+			if (!c->taken() && bt::GetCurrentTime() - i.data() > 5000)
 			{
 				if (c->getStatus() == Chunk::MMAPPED)
 					cache->save(c);
 				c->clear();
 				c->setStatus(Chunk::ON_DISK);
-				i = loaded.erase(i);
+				QMap<Uint32,TimeStamp>::iterator j = i;
+				i++;
+				loaded.erase(j);
 				num_removed++;
 			}
 			else
@@ -325,8 +341,8 @@ namespace bt
 				i++;
 			}
 		}
-//		Uint32 num_in_mem = loaded.count();
-		//Out() << QString("Cleaned %1 chunks, %2 still in memory").arg(num_removed).arg(num_in_mem) << endl;
+	//	Uint32 num_in_mem = loaded.count();
+	//	Out() << QString("Cleaned %1 chunks, %2 still in memory").arg(num_removed).arg(num_in_mem) << endl;
 	}
 	
 	void ChunkManager::saveChunk(unsigned int i,bool update_index)
@@ -343,6 +359,7 @@ namespace bt
 			if (update_index)
 			{
 				bitset.set(i,true);
+				todo.set(i,false);
 				recalc_chunks_left = true;
 				writeIndexFileEntry(c);
 				tor.updateFilePercentage(i,bitset);
@@ -350,6 +367,8 @@ namespace bt
 		}
 		else
 		{
+			c->clear();
+			c->setStatus(Chunk::NOT_DOWNLOADED);
 			Out(SYS_DIO|LOG_IMPORTANT) << "Warning: attempted to save a chunk which was excluded" << endl;
 		}
 	}
@@ -375,11 +394,39 @@ namespace bt
 		fptr.write(&hdr,sizeof(NewChunkHeader));
 	}
 	
+	Uint32 ChunkManager::onlySeedChunks() const
+	{
+		return only_seed_chunks.numOnBits();
+	}
+	
+	bool ChunkManager::completed() const
+	{
+		return todo.numOnBits() == 0;
+	}
+	
 	Uint64 ChunkManager::bytesLeft() const
 	{
-		Uint32 num_left = chunksLeft();
+		Uint32 num_left = bitset.getNumBits() - bitset.numOnBits();
 		Uint32 last = chunks.size() - 1;
-		if (!bitset.get(last) && !excluded_chunks.get(last))
+		if (last < chunks.size() && !bitset.get(last))
+		{
+			Chunk* c = chunks[last];
+			if (c)
+				return (num_left - 1)*tor.getChunkSize() + c->getSize();
+			else
+				return num_left*tor.getChunkSize();
+		}
+		else
+		{
+			return num_left*tor.getChunkSize();
+		}
+	}
+	
+	Uint64 ChunkManager::bytesLeftToDownload() const
+	{
+		Uint32 num_left = todo.numOnBits();
+		Uint32 last = chunks.size() - 1;
+		if (last < chunks.size() && todo.get(last))
 		{
 			Chunk* c = chunks[last];
 			if (c)
@@ -410,24 +457,47 @@ namespace bt
 		recalc_chunks_left = false;
 		return num;
 	}
+	
+	bool ChunkManager::haveAllChunks() const
+	{
+		return bitset.numOnBits() == bitset.getNumBits();
+	}
 
 	Uint64 ChunkManager::bytesExcluded() const
 	{
+		Uint64 excl = 0;
 		if (excluded_chunks.get(tor.getNumChunks() - 1))
 		{
 			Chunk* c = chunks[tor.getNumChunks() - 1];
 			Uint32 num = excluded_chunks.numOnBits() - 1;
-			return tor.getChunkSize() * num + c->getSize();
+			excl = tor.getChunkSize() * num + c->getSize();
 		}
 		else
 		{
-			return tor.getChunkSize() * excluded_chunks.numOnBits();
+			excl = tor.getChunkSize() * excluded_chunks.numOnBits();
 		}
+		
+		if (only_seed_chunks.get(tor.getNumChunks() - 1))
+		{
+			Chunk* c = chunks[tor.getNumChunks() - 1];
+			Uint32 num = only_seed_chunks.numOnBits() - 1;
+			excl += tor.getChunkSize() * num + c->getSize();
+		}
+		else
+		{
+			excl += tor.getChunkSize() * only_seed_chunks.numOnBits();
+		}
+		return excl;
 	}
 
 	Uint32 ChunkManager::chunksExcluded() const
 	{
-		return excluded_chunks.numOnBits();
+		return excluded_chunks.numOnBits() + only_seed_chunks.numOnBits();
+	}
+	
+	Uint32 ChunkManager::chunksDownloaded() const
+	{
+		return bitset.numOnBits();
 	}
 	
 	void ChunkManager::debugPrintMemUsage()
@@ -444,10 +514,27 @@ namespace bt
 		while (i <= to && i < chunks.count())
 		{
 			Chunk* c = chunks[i];
-			if(c->getPriority() != PREVIEW_PRIORITY || priority == EXCLUDED)
-				c->setPriority(priority);
+			c->setPriority(priority);
+			
+			if (priority == ONLY_SEED_PRIORITY)
+			{
+				only_seed_chunks.set(i,true);
+				todo.set(i,false);
+			}
+			else if (priority == EXCLUDED)
+			{
+				only_seed_chunks.set(i,false);
+				todo.set(i,false);
+			}
+			else
+			{
+				only_seed_chunks.set(i,false);
+				todo.set(i,!bitset.get(i));
+			}
+			
 			i++;
 		}
+		updateStats();
 	}
 
 	void ChunkManager::exclude(Uint32 from,Uint32 to)
@@ -461,6 +548,9 @@ namespace bt
 			Chunk* c = chunks[i];
 			c->setExclude(true);
 			excluded_chunks.set(i,true);
+			only_seed_chunks.set(i,false);
+			todo.set(i,false);
+			bitset.set(i,false);
 			i++;
 		}
 		recalc_chunks_left = true;
@@ -479,6 +569,8 @@ namespace bt
 			Chunk* c = chunks[i];
 			c->setExclude(false);
 			excluded_chunks.set(i,false);
+			if (!bitset.get(i))
+				todo.set(i,true);
 			i++;
 		}
 		recalc_chunks_left = true;
@@ -520,6 +612,9 @@ namespace bt
 	
 	void ChunkManager::loadFileInfo()
 	{
+		if (during_load)
+			return;
+		
 		File fptr;
 		if (!fptr.open(file_info_file,"rb"))
 			return;
@@ -551,6 +646,9 @@ namespace bt
 
 	void ChunkManager::savePriorityInfo()
 	{
+		if (during_load)
+			return;
+		
 		//save priority info and call saveFileInfo
 		saveFileInfo();
 		File fptr;
@@ -560,27 +658,35 @@ namespace bt
 			return;
 		}
 
-		QValueList<Uint32> dnd;
-		
-		Uint32 i = 0;
-		for ( ; i < tor.getNumFiles(); i++)
+		try
 		{
-			if(tor.getFile(i).getPriority() != NORMAL_PRIORITY)
+			QValueList<Uint32> dnd;
+			
+			Uint32 i = 0;
+			for ( ; i < tor.getNumFiles(); i++)
 			{
-				dnd.append(i);
-				dnd.append(tor.getFile(i).getPriority());
+				if(tor.getFile(i).getPriority() != NORMAL_PRIORITY)
+				{
+					dnd.append(i);
+					dnd.append(tor.getFile(i).getPriority());
+				}
 			}
-		}
-
-		Uint32 tmp = dnd.count();
-		fptr.write(&tmp,sizeof(Uint32));
-		// write all the non-default priority ones
-		for (i = 0;i < dnd.count();i++)
-		{
-			tmp = dnd[i];
+	
+			Uint32 tmp = dnd.count();
 			fptr.write(&tmp,sizeof(Uint32));
+			// write all the non-default priority ones
+			for (i = 0;i < dnd.count();i++)
+			{
+				tmp = dnd[i];
+				fptr.write(&tmp,sizeof(Uint32));
+			}
+			fptr.flush();
+		}
+		catch (bt::Error & err)
+		{
+			Out(SYS_DIO|LOG_IMPORTANT) << "Failed to save priority file " << err.toString() << endl;
+			bt::Delete(file_priority_file,true);
 		}
-		fptr.flush();
 	}
 	
 	void ChunkManager::loadPriorityInfo()
@@ -593,44 +699,58 @@ namespace bt
 			return;
 		}
 
-		Uint32 num = 0,tmp = 0;
+		Uint32 num = 0;
 		// first read the number of lines
-		if (fptr.read(&num,sizeof(Uint32)) != sizeof(Uint32))
+		if (fptr.read(&num,sizeof(Uint32)) != sizeof(Uint32) || num > 2*tor.getNumFiles())
 		{
 			Out(SYS_DIO|LOG_IMPORTANT) << "Warning : error reading chunk_info file" << endl;
 			loadFileInfo();
 			return;
 		}
 
+		Array<Uint32> buf(num);
+		if (fptr.read(buf,sizeof(Uint32)*num) != sizeof(Uint32)*num)
+		{
+			Out(SYS_DIO|LOG_IMPORTANT) << "Warning : error reading chunk_info file" << endl;
+			loadFileInfo();
+			return;
+		}
+		
+		fptr.close();
+		
 		for (Uint32 i = 0;i < num;i += 2)
 		{
-			if (fptr.read(&tmp,sizeof(Uint32)) != sizeof(Uint32))
-			{
-				Out(SYS_DIO|LOG_IMPORTANT) << "Warning : error reading chunk_info file" << endl;
-				loadFileInfo();
-				return;
-			}
-
-			bt::TorrentFile & tf = tor.getFile(tmp);
-			if (fptr.read(&tmp,sizeof(Uint32)) != sizeof(Uint32))
+			Uint32 idx = buf[i];
+			if (idx >= tor.getNumFiles())
 			{
 				Out(SYS_DIO|LOG_IMPORTANT) << "Warning : error reading chunk_info file" << endl;
 				loadFileInfo();
 				return;
 			}
 
+			bt::TorrentFile & tf = tor.getFile(idx);
+			
 			if (!tf.isNull())
 			{
-				switch(tmp)
+				// numbers are to be compatible with old chunk info files
+				switch(buf[i+1])
 				{
+				case FIRST_PRIORITY:
 				case 3:
 					tf.setPriority(FIRST_PRIORITY);
 					break;
+				case NORMAL_PRIORITY:
 				case 2:
 					tf.setPriority(NORMAL_PRIORITY);
 					break;
+				case EXCLUDED:
 				case 0:
-					tf.setDoNotDownload(true);
+					//tf.setDoNotDownload(true);
+					tf.setPriority(EXCLUDED);
+					break;
+				case ONLY_SEED_PRIORITY:
+				case -1:
+					tf.setPriority(ONLY_SEED_PRIORITY);
 					break;
 				default:
 					tf.setPriority(LAST_PRIORITY);
@@ -642,29 +762,20 @@ namespace bt
 
 	void ChunkManager::downloadStatusChanged(TorrentFile* tf,bool download)
 	{
-		savePriorityInfo();
 		Uint32 first = tf->getFirstChunk();
 		Uint32 last = tf->getLastChunk();
 		if (download)
 		{
-			// include the range of the file
-			if(tf->isMultimedia())
+			// include the chunks 
+			include(first,last);
+			
+			// if it is a multimedia file, prioritise first and last chunks of file
+			if (tf->isMultimedia())
 			{
 				prioritise(first,first+1,PREVIEW_PRIORITY);
 				if (last - first > 2)
 					prioritise(last -1,last, PREVIEW_PRIORITY);
 			}
-			if(chunks[first]->getPriority() > NORMAL_PRIORITY && first < tor.getNumChunks() - 1)
-				first++;
-			if(chunks[last]->getPriority() > NORMAL_PRIORITY && last > 0)
-				last--;
-			// last smaller then first is not normal, so just return
-			if (last < first)
-			{
-				cache->downloadStatusChanged(tf,download);
-				return;
-			}
-			include(first,last);
 		}
 		else
 		{
@@ -680,6 +791,7 @@ namespace bt
 			if (first == last && files.count() > 1)
 			{
 				cache->downloadStatusChanged(tf,download);
+				savePriorityInfo();
 				return;
 			}
 			
@@ -702,60 +814,97 @@ namespace bt
 				resetChunk(last);
 			}
 			
+			Priority maxp = ONLY_SEED_PRIORITY;
+			bool reprioritise_border_chunk = false;
+			bool modified = false;
+			
 			// if one file in the list needs to be downloaded,increment first
 			for (QValueList<Uint32>::iterator i = files.begin();i != files.end();i++)
 			{
-				if (!tor.getFile(*i).doNotDownload())
+				if (*i == tf->getIndex())
+					continue;
+				
+				const TorrentFile & other = tor.getFile(*i);
+				if (!other.doNotDownload())
 				{
-					if (first < tor.getNumChunks() - 1)
+					if (first != last && !modified)
+					{
 						first++;
-					break;
+						reprioritise_border_chunk = true;
+						modified = true;
+					}
+					
+					if (other.getPriority() > maxp)
+						maxp = other.getPriority();
 				}
 			}
 			
+			// in case we have incremented first, we better reprioritise the border chunk
+			if (reprioritise_border_chunk)
+				prioritise(first-1,first-1,maxp);
+			
+			maxp = ONLY_SEED_PRIORITY;
+			reprioritise_border_chunk = false;
+			modified = false;
+			
 			// if one file in the list needs to be downloaded,decrement last
 			for (QValueList<Uint32>::iterator i = last_files.begin();i != last_files.end();i++)
 			{
-				if (!tor.getFile(*i).doNotDownload())
+				if (*i == tf->getIndex())
+					continue;
+				
+				const TorrentFile & other = tor.getFile(*i);
+				if (!other.doNotDownload())
 				{
-					if (last > 0)
+					if (first != last && last > 0 && !modified)
+					{
 						last--;
-					break;
+						reprioritise_border_chunk = true;
+						modified = true;
+					}
+					
+					if (other.getPriority() > maxp)
+						maxp = other.getPriority();
 				}
 			}
+			
+			if (reprioritise_border_chunk)
+				prioritise(last+1,last+1,maxp);
 
 			// last smaller then first is not normal, so just return
 			if (last < first)
 			{
 				cache->downloadStatusChanged(tf,download);
+				savePriorityInfo();
 				return;
 			}
 			
 		//	Out(SYS_DIO|LOG_DEBUG) << "exclude " << first << " to " << last << endl;
 			exclude(first,last);
-			
 		}
 		// alert the cache but first put things in critical operation mode
 		cache->downloadStatusChanged(tf,download);
+		savePriorityInfo();
 	}
 
-	void ChunkManager::downloadPriorityChanged(TorrentFile* tf,Priority newpriority)
+	void ChunkManager::downloadPriorityChanged(TorrentFile* tf,Priority newpriority,Priority oldpriority)
 	{
-		if(newpriority == EXCLUDED)
+		if (newpriority == EXCLUDED)
 		{
 			downloadStatusChanged(tf, false);
 			return;
 		}
-		if(tf->getPriority() == EXCLUDED)
+		if (oldpriority == EXCLUDED)
 		{
 			downloadStatusChanged(tf, true);
 			return;
 		}
 
 		savePriorityInfo();
+		
 		Uint32 first = tf->getFirstChunk();
 		Uint32 last = tf->getLastChunk();
-
+		
 		// first and last chunk may be part of multiple files
 		// so we can't just exclude them
 		QValueList<Uint32> files;
@@ -765,17 +914,17 @@ namespace bt
 		
 		Chunk* c = chunks[first];
 		// if one file in the list needs to be downloaded,increment first
-		if (c->getPriority() == PREVIEW_PRIORITY)
-			first++;
-		else
+		for (QValueList<Uint32>::iterator i = files.begin();i != files.end();i++)
 		{
-			for (QValueList<Uint32>::iterator i = files.begin();i != files.end();i++)
+			Priority np = tor.getFile(*i).getPriority();
+			if (np > newpriority && *i != tf->getIndex())
 			{
-				if (tor.getFile(*i).getPriority() > newpriority && i != files.end())
-				{
-					first++;
-					break;
-				}
+				// make sure we don't go past last
+				if (first == last)
+					return;
+					
+				first++;
+				break;
 			}
 		}
 		
@@ -784,26 +933,30 @@ namespace bt
 		tor.calcChunkPos(last,files);
 		c = chunks[last];
 		// if one file in the list needs to be downloaded,decrement last
-		if (c->getPriority() == PREVIEW_PRIORITY)
-			last--;
-		else
+		for (QValueList<Uint32>::iterator i = files.begin();i != files.end();i++)
 		{
-			for (QValueList<Uint32>::iterator i = files.begin();i != files.end();i++)
+			Priority np = tor.getFile(*i).getPriority();
+			if (np > newpriority && *i != tf->getIndex())
 			{
-				if (tor.getFile(*i).getPriority() > newpriority && i != files.begin())
-				{
-					last--;
-					break;
-				}
+				// make sure we don't wrap around
+				if (last == 0 || last == first)
+					return;
+					
+				last--;
+				break;
 			}
 		}
+		
 		// last smaller then first is not normal, so just return
 		if (last < first)
 		{
 			return;
-		}			
+		}
+		
 
 		prioritise(first,last,newpriority);
+		if (newpriority == ONLY_SEED_PRIORITY)
+			excluded(first,last);
 	}
 	
 	bool ChunkManager::prepareChunk(Chunk* c,bool allways)
@@ -834,15 +987,17 @@ namespace bt
 			{
 				// We think we do not hae a chunk, but we do have it
 				bitset.set(i,true);
+				todo.set(i,false);
 				// the chunk must be on disk
 				c->setStatus(Chunk::ON_DISK);
 				tor.updateFilePercentage(i,bitset); 
 			}
 			else if (!ok_chunks.get(i) && bitset.get(i))
 			{
-				Out() << "Previously OK chunk " << i << " is corrupt !!!!!" << endl;
+				Out(SYS_DIO|LOG_IMPORTANT) << "Previously OK chunk " << i << " is corrupt !!!!!" << endl;
 				// We think we have a chunk, but we don't
 				bitset.set(i,false);
+				todo.set(i,!only_seed_chunks.get(i) && !excluded_chunks.get(i));
 				if (c->getStatus() == Chunk::ON_DISK)
 				{
 					c->setStatus(Chunk::NOT_DOWNLOADED);
@@ -861,6 +1016,7 @@ namespace bt
 		recalc_chunks_left = true;
 		saveIndexFile();
 		chunksLeft();
+		corrupted_count = 0;
 	}
 	
 	bool ChunkManager::hasExistingFiles() const
@@ -914,7 +1070,7 @@ namespace bt
 			tf.setMissing(false);
 			tf.setDoNotDownload(true); // set do not download
 		}
-		saveFileInfo();
+		savePriorityInfo();
 		saveIndexFile();
 		recalc_chunks_left = true;
 		chunksLeft();
